

#include <stdio.h>
#include <assert.h>

#include <set>
#include <algorithm>


#include "ndfa_dag.h"
#include "utils.h"



/****** NDFA_DAG public methods ******/

/** NDFA_DAG constructor
 *
 * This constructor produces an empty NDFA.
 */
NDFA_DAG :: NDFA_DAG () : NDFA_Node ( 0 )
{
    MESSAGE4 ( "Creating empty NDFA.\n" ) ;
    is_accept_all_ndfa = false ;
}

/** NDFA_DAG constructor
 *
 * This constructor produces an NDFA that accepts all bit patterns.
 * The accept state information is the one provided.
 */
NDFA_DAG :: NDFA_DAG ( Accept_State * asi ) : NDFA_Node ( 0 )
{
    MESSAGE4 ( "Creating accept-all NDFA.\n" ) ;

    NDFA_Node *     current_node ;

    current_node = this ;

    for ( int i = 0 ; i < BITS_PER_OPCODE ; i ++ )
    {
        accept_all_list [ i ] = current_node ;
        current_node = current_node -> Add_Transition ( true , true ) ;
    }

    current_node -> Make_Accept_State ( asi ) ;

    is_accept_all_ndfa = true ;
}



/** Reject_Pattern
 *
 * Makes the NDFA reject a particular bit pattern.
 */
void NDFA_DAG :: Reject_Pattern ( int start_bit , const char * pattern )
{
    MESSAGE4 ( "Rejecting bit pattern '%s' start %d.\n" ,
                        pattern , start_bit ) ;

    assert (( start_bit >= 0 )
            && ( start_bit < BITS_PER_OPCODE ) ) ;

    NDFA_Node *     current_node ;
    unsigned        i ;

    /* First add a series of transitions from the root to start_bit - 1,
     * that will accept any pattern. As a special case, if this is an
     * "accept-all" NDFA (created by the 2nd constructor), this set of
     * transitions already exists, so we can use an entry in the
     * accept_all_list. */

    if ( is_accept_all_ndfa )
    {
        current_node = accept_all_list [ start_bit ] ;
    } else {
        current_node = this ;

        for ( i = 0 ; i < (unsigned) start_bit ; i ++ )
        {
            current_node = current_node -> Add_Transition ( true , true ) ;
        }
    }

    /* Now add a series of transitions from start_bit - 1 to the end
     * of the pattern that reject that specific pattern */
    for ( i = 0 ; i < strlen ( pattern ) ; i ++ )
    {
        if ( pattern [ i ] == '0' )
        {
            current_node = current_node -> Add_Transition ( true , false ) ;
        } else if ( pattern [ i ] == '1' )
        {
            current_node = current_node -> Add_Transition ( false , true ) ;
        } else {
            assert ( 0 ) ;
        }
    }

    current_node -> Make_Reject_State () ;
}


/** Get_Accept_State 
 *
 * Translate an opcode number into an accept state.
 * If translation fails, 0 is returned.
 */
Accept_State * NDFA_DAG :: Get_Accept_State ( unsigned opcode )
{
    assert ( Is_Deterministic () ) ;

    NDFA_Node * node = this ;
    
    while (( node != 0L )
    && ( node -> Get_Accept_State () == 0L ))
    {
        if (( opcode >> (( BITS_PER_OPCODE - 1 ) - 
                    node -> Get_Test_Bit_Number () ) ) & 1 )
        {
            node = node -> DFA_Transition ( 1 ) ;
        } else {
            node = node -> DFA_Transition ( 0 ) ;
        }
    }
    if ( node == 0L )
    {
        return 0L ;
    } else {
        return node -> Get_Accept_State () ;
    }
}

/** Enable_Accept_State 
 *
 * Translate an opcode number into an accept state. During the tree
 * traversal, every tree node visited has it's "visit" counter incremented.
 * This tells the compression routines that this tree path will be needed.
 * The accept state's Enable method is called when it is reached.
 * If no accept state is found, 0 is returned.
 */
Accept_State * NDFA_DAG :: Enable_Accept_State ( unsigned opcode )
{
    assert ( Is_Deterministic () ) ;

    NDFA_Node * node = this ;
    
    while (( node != 0L )
    && ( node -> Get_Accept_State () == 0L ))
    {
        node -> Visit () ;

        if (( opcode >> (( BITS_PER_OPCODE - 1 ) - 
                    node -> Get_Test_Bit_Number () ) ) & 1 )
        {
            node = node -> DFA_Transition ( 1 ) ;
        } else {
            node = node -> DFA_Transition ( 0 ) ;
        }
    }
    if ( node == 0L )
    {
        return 0L ;
    } else {
        node -> Visit () ;
        node -> Get_Accept_State () -> Enable () ;
        return node -> Get_Accept_State () ;
    }
}


/** Generate_VHDL
 *
 * Produces VHDL, sent to the specified device, for the DFA,
 * which is first checked for determinism.
 */
void NDFA_DAG :: Generate_VHDL ( FILE * fd )
{
    assert ( Is_Deterministic () ) ;
    DFA_To_VHDL ( this , fd , "" ) ; 
}





/****** NDFA_DAG private methods ******/

void NDFA_DAG :: DFA_To_VHDL ( NDFA_Node * root_node , FILE * fd ,
                                    const char * indent )
{
    if ( root_node == 0L )
    {
        printf ( "Fault 1\n" ) ;
    }
    assert ( root_node != 0L ) ;

    char *  new_indent = new char [ strlen ( indent ) + 4 ] ;
    int     bit_num = ( BITS_PER_OPCODE - 1 ) - 
                        root_node -> Get_Test_Bit_Number () ;

    strcpy ( new_indent , indent ) ;
    strcat ( new_indent , "  " ) ;

    if ( root_node -> Get_Accept_State () != 0L )
    {
        root_node -> Get_Accept_State () -> 
                            Accept_State_To_VHDL ( fd , indent ) ;
    } else {
        /* Compressed DFA nodes always have two children */
        fprintf ( fd , "%sif ir ( %d ) = '1' then\n" , indent , bit_num ) ;
        DFA_To_VHDL ( root_node -> DFA_Transition ( 1 ) , fd , new_indent ) ;
        fprintf ( fd , "%selse --ir(%d)= '0'\n" , indent , bit_num ) ;
        DFA_To_VHDL ( root_node -> DFA_Transition ( 0 ) , fd , new_indent ) ;
        fprintf ( fd , "%send if ;\n" , indent ) ;
    }
    delete [] new_indent ;
}


