#include <stdio.h>

#include <sys/types.h>
#include <regex.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "opcode_map_reader.h"
#include "utils.h"
#include "definitions.h"
#include "ndfa_dag.h"
#include "ndfa_accept_state.h"
#include "state.h"
#include "state_machine.h"




/****** Opcode_Map_Reader public methods ******/



/** Read_Opcode_Map
 *
 * Read in an opcode map file, updating the ndfa structure.
 * Accept state information is obtained from New_Accept_State().
 */
int Opcode_Map_Reader :: Read_Opcode_Map ( const char * filename )
{
    const char *    read_regex_str = 
            "^(.... ... ... ......)  (.{8})(.{7})(.{16})(.{8})(.*)$" ;
    const int       NUMFIELDS = 6 ; /* number of regular expression fields */

    char            str [ MAX_LINE_LEN + 1 ] ;
    FILE *          fd = fopen ( filename , "rt" ) ;
    int             rc ;
    regex_t         read_regex ;
    int             opcodes_read = 0 ;
                    
    if ( fd == 0 )
    {
        return 0 ;
    }

    MESSAGE ( "Reading opcode map from %s\n" , filename ) ;

    rc = regcomp ( & read_regex , read_regex_str , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    /* read in a line from the map file */
    while ( fgets ( str , MAX_LINE_LEN , fd ) != NULL )
    {
        Remove_Trailing_Newlines ( str ) ;

        regmatch_t      matches [ NUMFIELDS + 1 ] ;

        if (( str [ 0 ] == '#' )     /* comment */
        || ( ! String_Contains_Non_Whitespace ( str ) ))    /* blank line */
        {
            continue ;
        }

        if ( regexec ( & read_regex , str , NUMFIELDS + 1 , 
                            matches , 0 ) != 0 )
        {
            /* line not recognised */
            throw new Unrecognised_Opcode_Map_Entry ( str ) ;
        }


        char *      micro_sub_name = 
                                Get_Regex_Match ( str , & matches [ 4 ] ) ;

        if ( ! String_Contains_Non_Whitespace ( micro_sub_name ) )
        {
            MESSAGE3 ( "Skipped '%s' - no micro sub name.\n" , str ) ;
            delete [] micro_sub_name ;
            continue ;
        }

        Remove_Whitespace_From_Ends ( micro_sub_name ) ;

        char *      bit_pattern = 
                                Get_Regex_Match ( str , & matches [ 1 ] ) ;
        char *      opcode_name = 
                                Get_Regex_Match ( str , & matches [ 2 ] ) ;
        char *      dea =       Get_Regex_Match ( str , & matches [ 3 ] ) ;
        char *      optimisation_info =   
                                Get_Regex_Match ( str , & matches [ 5 ] ) ;
        char *      comment =   Get_Regex_Match ( str , & matches [ 6 ] ) ;
        char        cmp_bit_pattern [ BITS_PER_OPCODE + 1 ] ;
        Accept_State * 
                    accept_state ;
        int         i , j ;

        Remove_Whitespace_From_Ends ( opcode_name ) ;
        Remove_Whitespace_From_Ends ( comment ) ;

        MESSAGE2 ( "Adding %s %s\n" , opcode_name , comment ) ;

        /* Remove all spaces from the bit pattern */
        for ( i = 0 , j = 0 ; i < (int) strlen ( bit_pattern ) ; i ++ )
        {
            if ( ! isspace ( bit_pattern [ i ] ) )
            {
                assert ( j < BITS_PER_OPCODE ) ;
                cmp_bit_pattern [ j ] = bit_pattern [ i ] ;
                j ++ ;
            }
        }
        assert ( j == BITS_PER_OPCODE ) ;
        cmp_bit_pattern [ j ] = '\0' ;

        /* Get an accept state for this line. */
        accept_state = New_Accept_State ( cmp_bit_pattern , opcode_name , 
                                            dea , micro_sub_name , 
                                            optimisation_info , comment ) ;
        assert ( accept_state != 0L ) ;

        /* Create a temporary NDFA that accepts any bit sequence
         * and has an accept state with information from "accept_state" */
        NDFA_DAG   temporary_ndfa ( accept_state ) ; 

        /* Begin parsing the bit pattern to determine which
         * patterns are NOT acceptable for this opcode. This
         * bit is highly specific to the format of the map file. */
        for ( i = 0 ; i < BITS_PER_OPCODE ; i ++ )
        {
            switch ( cmp_bit_pattern [ i ] )
            {
                case '0' :
                    temporary_ndfa . Reject_Pattern ( i , "1" ) ;
                    break ;
                case '1' :
                    temporary_ndfa . Reject_Pattern ( i , "0" ) ;
                    break ;
                case 'R' :
                case 'r' :
                case 'Z' :
                case 'D' :
                case 'e' :
                case 'a' :
                case 'I' :
                case 'X' :
                case 'C' :
                    /* These always match. They may be 1 or 0 */
                    break ;
                default :
                    if ( strncmp ( & cmp_bit_pattern [ i ] , 
                                                "SS" , 2 ) == 0 )
                    {
                        /* a size, that can be anything but 11 */
                        temporary_ndfa . Reject_Pattern ( i , "11" ) ;
                        i ++ ;
                    } else if ( strncmp ( & cmp_bit_pattern [ i ] , 
                                                "ss" , 2 ) == 0 )
                    {
                        /* another sort of size, that can be 
                         * anything but 00 */
                        temporary_ndfa . Reject_Pattern ( i , "00" ) ;
                        i ++ ;
                    } else if ( strncmp ( & cmp_bit_pattern [ i ] , 
                                                "TT" , 2 ) == 0 )
                    {
                        /* a bit operation type. Doesn't 
                        match if it is 00 */
                        temporary_ndfa . Reject_Pattern ( i , "00" ) ;
                        i ++ ;
                    } else if ( strncmp ( & cmp_bit_pattern [ i ] , 
                                                "cccc" , 4 ) == 0 )
                    {
                        /* Condition code.. may be anything apart
                        from 1. */
                        temporary_ndfa . Reject_Pattern ( i , "0001" ) ;
                        i += 3 ;
                    } else if ( strncmp ( & cmp_bit_pattern [ i ] , 
                                                "EEEEEE" , 6 ) == 0 )
                    {
                        /* An effective address field. */
                        Add_Reject_EAs ( & temporary_ndfa , i ,
                                        dea , false ) ;
                        i += 5 ;
                    } else if ( strncmp ( & cmp_bit_pattern [ i ] , 
                                                "MMMMMM" , 6 ) == 0 )
                    {
                        /* Another sort of effective address field. */
                        Add_Reject_EAs ( & temporary_ndfa , i , 
                                        dea , true ) ;
                        i += 5 ;
                    } else {
                        /* character not recognised */
                        assert ( 0 ) ;
                    }
                    break ;
            }
        }

        MESSAGE3 ( "  (NDFA size for '%s' = %d)\n" , 
                        opcode_name , temporary_ndfa . Get_Size () ) ;

        /* Turn the NDFA that accepts this opcode into a DFA.
         * The procedure assumes that any state that is both a
         * reject state and an accept state is just a reject state.
         * In this way, the resulting DFA accepts exactly the opcode
         * and rejects everything else. */
        temporary_ndfa . Make_Deterministic () ;

        MESSAGE3 ( "  (DFA size for '%s' = %d)\n" , 
                        opcode_name , temporary_ndfa . Get_Size () ) ;

        /* Can't compress the DFA here. It will compress to one accept
         * state! */

        /* And merge it into the main opcode 
         * We do not set any of the states in _this_ NDFA as REJECT,
         * (last parameter is false) because the reject states 
         * of one opcode may well be accept states of another. */
        ndfa . Merge_In ( & temporary_ndfa , false ) ;

        opcodes_read ++ ;

        MESSAGE3 ( "  (%d opcodes: master NDFA size is now %d)\n" , 
                        opcodes_read , ndfa . Get_Size () ) ;

        /* Free up memory used to store regex fields */
        delete [] bit_pattern ;
        delete [] opcode_name ;
        delete [] dea ;
        delete [] micro_sub_name ;
        delete [] comment ;
    }
    regfree ( & read_regex ) ; 

    MESSAGE ( "Making decoding DFA..\n" ) ;

    ndfa . Make_Deterministic () ;

    MESSAGE ( "Decoding DFA has %d nodes.\n" , ndfa . Get_Size () ) ;

    return opcodes_read ;
}




/****** Opcode_Map_Reader private methods ******/

/** Add_Reject_EAs
 *
 * A helper function for adding a series of rejection states to
 * an NDFA, to represent an effective address field.
 */
void Opcode_Map_Reader :: Add_Reject_EAs ( NDFA_DAG * n , int offset ,
                const char * dea , bool move_destination_ea )
{
    /* parse the list of disallowed address modes */
    if ( index ( dea , 'i' ) != NULL )
    {
        /* Immediates are not allowed. */
        n -> Reject_Pattern ( offset , 
            move_destination_ea ? "100111" : "111100" ) ;
    }
    if ( index ( dea , 'p' ) != NULL )
    {
        /* PC-relative is not allowed.
         not indirect with displacement: */
        n -> Reject_Pattern ( offset , 
            move_destination_ea ? "010111" : "111010" ) ;
        /* nor memory indirect with index */
        n -> Reject_Pattern ( offset , 
            move_destination_ea ? "011111" : "111011" ) ;
    }
    if ( index ( dea , 'a' ) != NULL )
    {
        /* Address Register Direct: not allowed */
        n -> Reject_Pattern (
            move_destination_ea ? ( offset + 3 ) : offset , "001" ) ;
    }
    if ( index ( dea , 'd' ) != NULL )
    {
        /* Data Register Direct: not allowed */
        n -> Reject_Pattern ( 
            move_destination_ea ? ( offset + 3 ) : offset , "000" ) ;
    }
    if (( index ( dea , 'r' ) != NULL )
    || ( index ( dea , '+' ) != NULL ))
    {
        /* Register modifying modes not allowed...
        no postinc */
        n -> Reject_Pattern (
            move_destination_ea ? ( offset + 3 ) : offset , "011" ) ;
    }
    if (( index ( dea , 'r' ) != NULL )
    || ( index ( dea , '-' ) != NULL ))
    {
        /* Register modifying modes not allowed...
        no predec */
        n -> Reject_Pattern (
            move_destination_ea ? ( offset + 3 ) : offset , "100" ) ;
    }

    /* These three modes are reserved by Motorola
    and are thus never valid. */
    n -> Reject_Pattern ( offset ,
        move_destination_ea ? "101111" : "111101" ) ;
    n -> Reject_Pattern ( offset ,
        move_destination_ea ? "110111" : "111110" ) ;
    n -> Reject_Pattern ( offset ,
        move_destination_ea ? "111111" : "111111" ) ;
}


