
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <regex.h>
#include <assert.h>

#include "settings.h"
#include "utils.h"
#include "exceptions.h"
#include "definitions.h"



/** Settings constructor
 *
 * Construct a new object to read a settings ini file.
 */
Settings :: Settings ( const char * ini_file )
{
    const char *    read_regex_str = "^(.+[^ ]) *= *([^ ].*)$" ;
    const int       NUMFIELDS = 2 ; /* number of regular expression fields */

    const char *    verbose_level_str ;

    const Variable_Table var_table [] = { 
        { "output filename" , & output_filename , true } ,
        { "opcode map filename" , & opcode_map_filename , true } ,
        { "state machine directory" , & state_machine_directory , true } ,
        { "root vhdl input filename" , & root_vhdl_input_filename , true } ,
        { "vhdl input directory" , & vhdl_input_directory , true } ,
        { "initial state machine" , & initial_state_machine , true } ,
        { "required opcode file" , & required_opcode_file , false } ,
        { "verbose level" , & verbose_level_str , false } ,
        { 0L , 0L , false } } ;

    FILE *                  fd = fopen ( ini_file , "rt" ) ;
    char                    str [ MAX_LINE_LEN + 1 ] ;
    int                     rc ;
    regex_t                 read_regex ;
    const Variable_Table *  table_ptr = & var_table [ 0 ] ;
    

    if ( fd == 0L )
    {
        throw new File_Access_Exception ( ini_file ) ;
    }

    /* Compile the regular expression used for reading */
    rc = regcomp ( & read_regex , read_regex_str , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    /* Zero the configuration string pointers */
    table_ptr = & var_table [ 0 ] ;
    while ( table_ptr -> var_name != 0L ) 
    {
        (* table_ptr -> var_ptr) = 0L ;
        table_ptr ++ ;
    }

    MESSAGE3 ( "Reading settings from %s\n" , ini_file ) ;

    while ( fgets ( str , MAX_LINE_LEN , fd ) != NULL )
    {
        Remove_Trailing_Newlines ( str ) ;
        Remove_Whitespace_From_Ends ( str ) ;

        if (( str [ 0 ] == '#' )     /* comment */
        || ( ! String_Contains_Non_Whitespace ( str ) ))    /* blank line */
        {
            continue ;
        }

        regmatch_t      matches [ NUMFIELDS + 1 ] ;

        if ( regexec ( & read_regex , str , NUMFIELDS + 1 , 
                            matches , 0 ) != 0 )
        {
            /* line not recognised */
            throw new Unrecognised_Ini_File_Entry ( str ) ;
        }

        const char * variable_name = Get_Regex_Match ( str , & matches [ 1 ] ) ;
        const char * value = Get_Regex_Match ( str , & matches [ 2 ] ) ;


        if ( strlen ( value ) == 0 )
        {
            throw new Unrecognised_Ini_File_Entry ( str ) ;
        }


        /* Now, what variable is this? */

        bool    found = false ;

        table_ptr = & var_table [ 0 ] ;
        while ( table_ptr -> var_name != 0L ) 
        {
            if ( strcasecmp ( table_ptr -> var_name , variable_name ) == 0 )
            {
                /* The variable has been found. Set the value.
                 * If the value has already been assigned, zero it. */
                if ( (* table_ptr -> var_ptr) != 0L )
                {
                    delete [] (* table_ptr -> var_ptr) ;
                }
                (* table_ptr -> var_ptr) = value ;
                found = true ;
            }
            table_ptr ++ ;
        }

        if ( ! found )
        {
            /* The variable wasn't found. */
            throw new Unrecognised_Ini_File_Entry ( str ) ;
        }
        delete [] variable_name ;
        /* Don't free the value string - it has been assigned to one
         * of the variables. */
    }

    /* Were all the required variables assigned? 
     * While we check this, print out a list of the settings.
     */

    MESSAGE2 ( "Settings list:\n" ) ;

    table_ptr = & var_table [ 0 ] ;
    while ( table_ptr -> var_name != 0L ) 
    {

        if ((* table_ptr -> var_ptr) != 0L )
        {
            MESSAGE2 ( " %s = %s\n" , table_ptr -> var_name ,
                                (* table_ptr -> var_ptr) ) ;
        } else {
            if ( table_ptr -> required )
            {
                throw new No_Setting_For_Variable ( table_ptr -> var_name ) ;
            }
        }
        table_ptr ++ ;
    }

    regfree ( & read_regex ) ;

    /* Was the verbose level set? */

    if ( verbose_level_str != 0L )
    {
        verbose_level = atoi ( verbose_level_str ) ;
        if ( verbose_level < 1 )
        {
            verbose_level = 1 ;
        }
        delete [] verbose_level_str ;
    } else {
        verbose_level = VERBOSE_LOW ;
    }
}

/** Settings destructor
 */
Settings :: ~Settings ()
{
    delete [] output_filename ;
    delete [] opcode_map_filename ;
    delete [] state_machine_directory ;
    delete [] root_vhdl_input_filename ;
    delete [] vhdl_input_directory ;
    delete [] initial_state_machine ;

    if ( required_opcode_file != 0L )
    {
        delete [] required_opcode_file ;
    }
}

/** Set_Required_Opcode_File
 *
 * Allows the caller to change/set the required opcode file name.
 */
void Settings :: Set_Required_Opcode_File ( const char * fn )
{
    if ( required_opcode_file != 0L )
    {
        delete [] required_opcode_file ;
    }
    required_opcode_file = Copy_String ( fn ) ;
}

/** Get_Required_Opcode_File
 *
 * Returns the required opcode filename, or throws an exception if
 * it is unset.
 */
const char * Settings :: Get_Required_Opcode_File ()
{
    if ( required_opcode_file == 0L )
    {
        throw new No_Setting_For_Variable ( "required opcode file" ) ;
    }
    return required_opcode_file ;
}



